unit logcont;

{$mode objfpc}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics,
  Dialogs, StrUtils, dateutils,
  lazutf8sysutils, SynMemo, //Defines NOWUTC.
  fileview, //For opening log files.
  dlheader, //For Timezone conversions.
  appsettings, //For saving and restoring hotkeys.
  StdCtrls, ExtCtrls, Buttons, ValEdit, Grids, header_utils;

type

  { TFormLogCont }

  TFormLogCont = class(TForm)
    BitBtn1: TBitBtn;
    AnnotateButton: TButton;
    AnnotateEdit: TEdit;
    AnnotationGroupBox: TGroupBox;
    RecordsMissedLabel: TLabel;
    RecordsViewLabel: TLabel;
    LogFileNameText: TEdit;
    LogfileNameLabel: TLabel;
    Labelinfile: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Labelfile: TLabel;
    LabelIn: TLabel;
    Label2: TLabel;
    LCTrigSeconds: TLabeledEdit;
    LCTrigMinutes: TLabeledEdit;
    NextRecordAt: TStaticText;
    CurrentTime: TStaticText;
    NextRecordIn: TStaticText;
    RecordsLogged: TStaticText;
    FilesLogged: TStaticText;
    CloseButton: TButton;
    OpenLogDialog: TOpenDialog;
    RecordsViewMemo: TSynMemo;
    StartButton: TBitBtn;
    LCTriggerGroup: TRadioGroup;
    RecordsMissed: TStaticText;
    StopButton: TBitBtn;
    FineTimer: TTimer;
    RecordsViewSynMemo: TSynMemo;
    StringGrid1: TStringGrid;
    procedure AnnotateButtonClick(Sender: TObject);
    procedure AnnotateEditKeyPress(Sender: TObject; var Key: char);
    procedure BitBtn1Click(Sender: TObject);
    procedure CloseButtonClick(Sender: TObject);
    procedure FineTimerTimer(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: word; Shift: TShiftState);
    procedure FormShow(Sender: TObject);
    procedure StartButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure StringGrid1KeyUp(Sender: TObject; var Key: word; Shift: TShiftState);
    procedure StringGrid1SelectEditor(Sender: TObject; aCol, aRow: integer;
      var Editor: TWinControl);
  private
    { private declarations }
  public
    { public declarations }
    procedure CustomizeEdit(ACol, ARow: integer; Editor: TStringCellEditor);
  end;

var
  FormLogCont: TFormLogCont;
  LogTimePreset, LogTimeCurrent: integer;
  LCLoggedCount: integer; // Number of logged records taken during this session
  LCLogFileCount: integer; //Number of logfiles created during this session
  DesiredLogTime: TDateTime;
  StopRecording: boolean;
  Setting: string; //Setting of recorder
  Recording: boolean; //Flag to indicate that record is being logged
  RecordingMode: boolean = False; //Indicate that user has pressed Record button
  TimerBusy: boolean = False; //Flag to indicate that Timer event is being handled

  //Used to check 24 hr log file rollover
  LCStartFileTime: TDateTime;

  RecordsMissedCount: integer;
  //Determined while recording records, incrememts if meter does not respond.

  OldSecond: word; //Previous second value used to check timer rollover.

  AnnotateText: string;

implementation

uses Unit1;

function Sec2DHMS(InSeconds: integer): string;
var
  Days: integer;
  Hours: integer;
  InHours: integer;
  Minutes: integer;
  InMinutes: integer;
  Seconds: integer;
begin
  Seconds := InSeconds mod 60;
  InMinutes := InSeconds div 60;

  Minutes := InMinutes mod 60;
  InHours := InMinutes div 60;

  Hours := InHours mod 24;
  Days := InHours div 24;

  Sec2DHMS := Format('%.1dd %.2d:%.2d:%.2d', [Days, Hours, Minutes, Seconds]);
end;

procedure OnTheClock(const ThisMoment: TDateTime; const Granularity: integer);
{This utility gets the next time from now based on a granularity setting,
 i.e for determining log time and difference for logging every x minutes}
begin
  // Determine new record time and calculate time until record
  DesiredLogTime := ThisMoment; //Copy current time for manipulation
  DesiredLogTime := IncMinute(DesiredLogTime, Granularity -
    MinuteOf(ThisMoment) mod Granularity);
  //Replace seconds with 0
  DesiredLogTime := RecodeSecond(DesiredLogTime, 0);
  FormLogCont.NextRecordAt.Caption :=
    FormatDateTime('yyyy-mm-dd hh:nn:ss', DesiredLogTime);
  LogTimeCurrent := SecondsBetween(ThisMoment, DesiredLogTime);
end;

procedure LogOneReading();
var
  pieces: TStringList;
  Result: string;
  ThisMoment, ThisMomentUTC: TDateTime;
  subfix: ansistring;

  procedure WriteRecord(Special: string = '');
  var
    ComposeString: string;
  begin
    ThisMomentUTC := NowUTC;
    ThisMoment := dlheader.ptz.GMTToLocalTime(ThisMomentUTC,
      DLHeaderForm.TZLocationBox.Text, subfix);

    Application.ProcessMessages;

    //Create new logfile if over 24hr UTC time.
    if (DayOf(LCStartFileTime) <> DayOf(ThisMoment)) then
    begin
      // Open new file and store header
      WriteDLHeader('LE', Format('Logged continuously %s.', [setting]));

      LCStartFileTime := ThisMoment;

      // Display number of logfiles
      Inc(LCLogFileCount);
      FormLogCont.FilesLogged.Caption := IntToStr(LCLogFileCount);
      if LCLogFileCount = 1 then
        FormLogCont.Labelfile.Caption := 'file'
      else
        FormLogCont.Labelfile.Caption := 'files';
    end
    else
    begin
      AssignFile(DLRecFile, LogFileName);
      Append(DLRecFile); //Open file for appending
    end;

    // Display LogFileNameText path
    FormLogCont.LogFileNameText.Caption := LogFileName;

    if (Special = '') then
    begin
      // YYYY-MM-DDTHH:mm:ss.fff;YYYY-MM-DDTHH:mm:ss.fff;Celsius;number;Hz;mag/arcsec^2')
      ComposeString :=
        FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"', ThisMomentUTC) //Date UTC
        //+ FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"', Now) //Date Local
        //+ FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"', DesiredLogTime)
        + FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"',
        ThisMoment)  //Date Local (calculated)
        + Format('%1.1f;', [StrToFloatDef(
        AnsiLeftStr(pieces.Strings[5], length(pieces.Strings[5]) - 1),
        0, FPointSeparator)])
        //Temperature
        + Format('%d;', [StrToIntDef(
        AnsiLeftStr(pieces.Strings[3], length(pieces.Strings[3]) - 1), 0)]) //counts
        + Format('%d;', [StrToIntDef(
        AnsiLeftStr(pieces.Strings[2], length(pieces.Strings[2]) - 2), 0)]) //Hz
        + Format('%1.2f', [StrToFloatDef(
        AnsiLeftStr(pieces.Strings[1], length(pieces.Strings[1]) - 1),
        0, FPointSeparator)]) //mpsas value
      ;

      if (Length(AnnotateText) > 0) then
      begin
        ComposeString := ComposeString + ';' + AnnotateText;
        AnnotateText := '';
      end;

    end
    else
    begin
      ComposeString :=
        FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"', NowUTC) //Date UTC
        + FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz";"', Now) //Date Local
        + Special;
    end;

    Writeln(DLRecFile, ComposeString);
    FormLogCont.RecordsViewSynMemo.Append(ComposeString);
    //Ensure that latest append is displayed at bottom of window.
    FormLogCont.RecordsViewSynMemo.CaretY := FormLogCont.RecordsViewSynMemo.Lines.Count;
    Inc(LCLoggedCount);
    FormLogCont.RecordsLogged.Caption := IntToStr(LCLoggedCount);

  end;

begin
  Recording := True;
  pieces := TStringList.Create;
  pieces.Delimiter := ',';
  pieces.StrictDelimiter := False; //Parse spaces also

  //Try a first time
  Result := GetReading;
  pieces.DelimitedText := Result;
  if (pieces.Count = 6) then
    WriteRecord
  else
  begin //Try a second time
    Result := GetReading;
    pieces.DelimitedText := Result;
    if (pieces.Count = 6) then
      WriteRecord
    else
    begin //Try a second time
      Result := GetReading;
      pieces.DelimitedText := Result;
      if (pieces.Count = 6) then
        WriteRecord
      else
      begin
        Inc(RecordsMissedCount);
        WriteRecord(';;;'); //Empty fields
        FormLogCont.RecordsMissed.Color := clRed;
        FormLogCont.RecordsMissed.Font.Color:=clWhite;
        FormLogCont.RecordsMissed.Caption := IntToStr(RecordsMissedCount);
      end;
    end;
  end;

  Flush(DLRecFile);
  Close(DLRecFile);
  Recording := False;
end;

{ TFormLogCont }

procedure TFormLogCont.BitBtn1Click(Sender: TObject);
begin
  OpenLogDialog.InitialDir := appsettings.LogsDirectory;
  if OpenLogDialog.Execute then
  begin
    Form2.Memo1.Lines.LoadFromFile(OpenLogDialog.Filename);
    Form2.Show;
  end;

end;

procedure TFormLogCont.AnnotateButtonClick(Sender: TObject);
begin
  //Initiate a recording with annotated text while already in record-mode.
  AnnotateText := FormLogCont.AnnotateEdit.Text;
  LogOneReading;
end;

procedure TFormLogCont.AnnotateEditKeyPress(Sender: TObject; var Key: char);
begin
  // Only annotate if recording.
  // Annotate when the enter key (#13) is pressed.
  if ((RecordingMode) and (Key = #13)) then
    AnnotateButtonClick(Sender);
end;

procedure TFormLogCont.CloseButtonClick(Sender: TObject);
begin
  Close;
end;

procedure TFormLogCont.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  if CloseButton.Enabled then
    Close
  else
    StopRecording := True;

end;

procedure TFormLogCont.FormCreate(Sender: TObject);
begin

end;

procedure TFormLogCont.FormKeyUp(Sender: TObject; var Key: word; Shift: TShiftState);
var
  AnnotateString: string;
  i: integer;

begin
  if RecordingMode then
  begin
    //Search for key in hotkey list
    for i := 1 to StringGrid1.RowCount - 1 do
    begin
      if (char(Key) = vConfigurations.ReadString('HotKeys', 'h' +
        IntToStr(i - 1), '')) then
      begin
        AnnotateText := vConfigurations.ReadString('HotKeys', 'a' +
          IntToStr(i - 1), '');
        LogOneReading;
      end;
    end;
  end;
end;

procedure TFormLogCont.FormShow(Sender: TObject);
var
  i: integer;
begin
  RecordingMode := False;
  StopButton.Enabled := False;
  StartButton.Enabled := True;
  LogTimePreset := 0;
  LogTimeCurrent := 0;
  NextRecordIn.Caption := '';
  LCLoggedCount := 0;
  RecordsLogged.Caption := IntToStr(LCLoggedCount);
  LCLogFileCount := 0;
  FilesLogged.Caption := IntToStr(LCLogFileCount);
  OldSecond := SecondOf(Now);
  RecordsMissedCount := 0;
  RecordsMissed.Caption := IntToStr(RecordsMissedCount);
  AnnotateButton.Enabled := False;


  for i := 1 to StringGrid1.RowCount - 1 do
  begin
    StringGrid1.Cells[0, i] :=
      vConfigurations.ReadString('HotKeys', 'h' + IntToStr(i - 1), '');
    StringGrid1.Cells[1, i] :=
      vConfigurations.ReadString('HotKeys', 'a' + IntToStr(i - 1), '');
  end;

end;

procedure TFormLogCont.FineTimerTimer(Sender: TObject);
var
  ThisMoment: TDateTime;
begin
  // Only trigger out once per second with ~20ms accuracy.
  // This prevents general drift in recordings.

  FineTimer.Enabled := False;

  if (SecondOf(Now) <> OldSecond) then
  begin
    OldSecond := SecondOf(Now);
    if (RecordingMode) then
    begin
      //Gets triggered by the fine timer once per second if RecordingMode is true.

      if (Recording = False) then
        //Prevent more recording while already busy saving a record
      begin
        ThisMoment := RecodeMilliSecond(Now, 0);
        CurrentTime.Caption := FormatDateTime('yyyy-mm-dd hh:nn:ss', ThisMoment);

        // Check if counter has expired while in Seconds or Minutes mode
        case LCTriggerGroup.ItemIndex of
          0..1:
          begin // Seconds or Minutes mode selected
            Dec(LogTimeCurrent);
            if (LogTimeCurrent <= 0) then
            begin
              //LogContTimer.Enabled := False;

              // Log value//* threshold checking
              LogOneReading;

              // Restart counter for continuous logging
              LogTimeCurrent := LogTimePreset;
              NextRecordAt.Caption :=
                FormatDateTime('yyyy-mm-dd hh:nn:ss',
                IncSecond(ThisMoment, LogTimeCurrent));
              NextRecordIn.Caption := Sec2DHMS(LogTimeCurrent);

              RecordsLogged.Caption := IntToStr(LCLoggedCount);
              //LogContTimer.Enabled := True;
            end
            else
              NextRecordIn.Caption := Sec2DHMS(LogTimeCurrent);

          end;

          2..7:
          begin // On-the-clock mode selected
            if (CompareDateTime(ThisMoment, DesiredLogTime) >= 0) then
            begin
              //LogContTimer.Enabled := False;

              // Log value//* threshold checking
              LogOneReading;

              case LCTriggerGroup.ItemIndex of
                2: OnTheClock(ThisMoment, 1); //Every 1 minute
                3: OnTheClock(ThisMoment, 5); //Every 5 minutes
                4: OnTheClock(ThisMoment, 10); //Every 10 minutes
                5: OnTheClock(ThisMoment, 15); //Every 15 minutes
                6: OnTheClock(ThisMoment, 30); //Every 30 minutes
                76: OnTheClock(ThisMoment, 60); //Every hour
              end;

              RecordsLogged.Caption := IntToStr(LCLoggedCount);
              //LogContTimer.Enabled := True;
            end
            else //continue counting
            begin
              LogTimeCurrent := Round(SecondSpan(ThisMoment, DesiredLogTime));
              NextRecordIn.Caption := Sec2DHMS(LogTimeCurrent);
            end;
          end;
        end;

        if StopRecording then
        begin
          RecordingMode := False;
          StopButton.Enabled := False;
          StartButton.Enabled := True;
          LCTrigSeconds.Enabled := True;
          LCTrigMinutes.Enabled := True;
          LogTimePreset := 0;
          LogTimeCurrent := 0;
          NextRecordIn.Caption := '';

          CloseButton.Enabled := True;
          LCTriggerGroup.Enabled := True;
        end;
      end; //End of checking Recording flag

    end;

  end;

  //Set length of timer and start it if still recording
  if not StopRecording then
  begin
    FineTimer.Interval := 1000 - MilliSecondOf(Now);
    FineTimer.Enabled := True;
  end;

end;


procedure TFormLogCont.StartButtonClick(Sender: TObject);  //Start recording
var
  ThisMoment: TDateTime;
  LogMinutes: integer;
begin
  StopRecording := False;

  //Grab selected time, convert to seconds, store in LogTimePreset

  //Grab current time:
  ThisMoment := RecodeMilliSecond(Now, 0); //Strip fractional seconds for display
  CurrentTime.Caption := FormatDateTime('yyyy-mm-dd hh:nn:ss', ThisMoment);

  case LCTriggerGroup.ItemIndex of
    0: //Every x seconds (on the second)
    begin
      LogTimePreset := StrToIntDef(LCTrigSeconds.Text, 1);
      LCTrigSeconds.Text := IntToStr(LogTimePreset);
      LogTimeCurrent := LogTimePreset;
      NextRecordAt.Caption := FormatDateTime('yyyy-mm-dd hh:nn:ss',
        IncSecond(ThisMoment, LogTimeCurrent));
      Setting := Format('every %d seconds', [LogTimePreset]);
    end;
    1: //Every x minutes (at 0 seconds)
    begin
      { TODO : ontheclock x minutes }
      LogMinutes := StrToIntDef(LCTrigMinutes.Text, 1);
      LogTimePreset := LogMinutes * 60;
      LCTrigMinutes.Text := IntToStr(LogTimePreset div 60);
      LogTimeCurrent := LogTimePreset;
      NextRecordAt.Caption := FormatDateTime('yyyy-mm-dd hh:nn:ss',
        IncSecond(ThisMoment, LogTimeCurrent));
      if LogMinutes = 1 then
        Setting := Format('every 1 minute', [LogMinutes])
      else
        Setting := Format('every %d minutes', [LogMinutes]);
    end;
    2: //Every 1 minutes
    begin
      OnTheClock(ThisMoment, 1);
      Setting := 'every 1 minute on the minute';
    end;
    3: //Every 5 minutes
    begin
      OnTheClock(ThisMoment, 5);
      Setting := 'every 5 minutes on the 1/12th hour';
    end;
    4://Every 10 minutes
    begin
      OnTheClock(ThisMoment, 10);
      Setting := 'every 10 minutes on the 1/6th hour';
    end;
    5://Every 15 minutes
    begin
      OnTheClock(ThisMoment, 15);
      Setting := 'every 15 minutes on the 1/4 hour';
    end;
    6://Every 30 minutes
    begin
      OnTheClock(ThisMoment, 30);
      Setting := 'every 30 minutes on the 1/2 hour';
    end;
    7://Every hour
    begin
      OnTheClock(ThisMoment, 60);
      Setting := 'every hour on the hour';
    end;
  end;

  NextRecordIn.Caption := Sec2DHMS(LogTimeCurrent);
  RecordingMode := True;
  StopButton.Enabled := True;
  StartButton.Enabled := False;
  CloseButton.Enabled := False;
  LCTrigSeconds.Enabled := False;
  LCTrigMinutes.Enabled := False;
  LCLoggedCount := 0;
  RecordsLogged.Caption := IntToStr(LCLoggedCount);
  LCTriggerGroup.Enabled := False;
  RecordsViewSynMemo.Clear;
  AnnotateButton.Enabled := True;
  RecordsMissed.Color := clDefault;
  RecordsMissed.Font.Color := clDefault;
  RecordsMissedCount := 0;
  RecordsMissed.Caption := IntToStr(RecordsMissedCount);

  try
    // Open file and store header
    WriteDLHeader('LE', Format('Logged continuously %s.', [setting]));
    Flush(DLRecFile);
    CloseFile(DLRecFile);
    LCLogFileCount := 1;
    FilesLogged.Caption := IntToStr(LCLogFileCount);
    Labelfile.Caption := 'file';

    //Save initial log time for 24 rollover check later
    LCStartFileTime := ThisMoment;

    LogOneReading; //First reading

  except
    StatusMessage('ERROR! IORESULT: ' + IntToStr(IOResult));
  end;

  //Set length of timer and start it
  FineTimer.Interval := 1000 - MilliSecondOf(Now);
  FineTimer.Enabled := True;

end;

procedure TFormLogCont.StopButtonClick(Sender: TObject);
begin
  StopRecording := True;
end;

procedure TFormLogCont.StringGrid1KeyUp(Sender: TObject; var Key: word;
  Shift: TShiftState);
var
  i, Count: integer;

  function StrippedOfNonAscii(const s: string): string;
  var
    i, Count: integer;
  begin
    SetLength(Result, Length(s));
    Count := 0;
    for i := 1 to Length(s) do
    begin
      if ((s[i] >= #32) and (s[i] < #127)) or (s[i] in [#10, #13]) then
      begin
        Inc(Count);
        Result[Count] := s[i];
      end;
    end;
    SetLength(Result, Count);
  end;

begin
  for i := 1 to StringGrid1.RowCount - 1 do
  begin
    vConfigurations.WriteString('HotKeys', 'h' + IntToStr(i - 1),
      StringGrid1.Cells[0, i]);
    vConfigurations.WriteString('HotKeys', 'a' + IntToStr(i - 1),
      StrippedOfNonAscii(StringGrid1.Cells[1, i]));
  end;
end;

procedure TFormLogCont.StringGrid1SelectEditor(Sender: TObject;
  aCol, aRow: integer; var Editor: TWinControl);
begin
  if Editor is TStringCellEditor then
    CustomizeEdit(aCol, aRow, Editor as TStringCellEditor);
end;

procedure TFormLogCont.CustomizeEdit(ACol, ARow: integer; Editor: TStringCellEditor);
begin
  if aCol = 0 then
    Editor.MaxLength := 1
  else
    Editor.MaxLength := 0;

end;

initialization
  {$I logcont.lrs}

end.
